"""
Module MDAV (generic)
"""

from collections import Counter
from statistics import mean, median_low, stdev

from logzero import logger

from mdav.common import Column, Dataset, MDAVType


def calc_mdav_generic(
        dataset: Dataset,
        k: int) -> list:
    """
    Function calc_mdav_generic
    """
    # region Auxiliary functions
    def get_columns(dataset):
        columns = []
        for i in range(len(dataset.records[0])):
            column_record = []
            for record in dataset.records:
                column_record += [record[i]]
            columns.append(Column(column_record, dataset.types[i]))
        return columns

    def compute_means_and_stdevs(columns):
        means = []
        stdevs = []
        for i in range(len(columns)):
            if columns[i].type == MDAVType.continous:
                means.append(mean(columns[i].record))
                stdevs.append(stdev(columns[i].record))
            elif columns[i].type == MDAVType.ordinal:
                means.append(len(set(columns[i].record)))
                stdevs.append(None)
            else:
                means.append(None)
                stdevs.append(None)
        return means, stdevs

    def compute_auxiliary_tables(dataset: Dataset):
        # TODO: This code is highly inefficient!!
        # Compute auxiliary values
        columns = get_columns(dataset)
        means, stdevs = compute_means_and_stdevs(columns)
        return columns, means, stdevs

    # endregion

    # region Main functions to calculate averages, distances and clusters

    def get_distance(x: list, y: list) -> float:
        """
        Function get_distance
        """
        def get_distance_column(x, y):
            def get_distance_continous(x, y):
                x_normalized = (x - means[i])/stdevs[i]
                y_normalized = (y - means[i])/stdevs[i]
                return abs(x_normalized - y_normalized)

            def get_distance_ordinal(x, y):
                # TODO: This assumes that every value is a category
                if means[i] != 0:
                    return abs(x - y) / means[i]
                else:
                    return 0

            def get_distance_nominal(x, y):
                return 0 if x == y else 1

            if dataset.types[i] == MDAVType.continous:
                distance = get_distance_continous(x, y)
            elif dataset.types[i] == MDAVType.ordinal:
                distance = get_distance_ordinal(x, y)
            elif dataset.types[i] == MDAVType.nominal:
                distance = get_distance_nominal(x, y)
            elif dataset.types[i] == MDAVType.ignore:
                distance = ''  # This should not happen

            return distance

        distance_values = []
        for i in range(len(x)):
            distance_values.append(get_distance_column(x[i], y[i]))

        # TODO: Is this right?
        return mean(distance_values)

    def compute_average_record(dataset: Dataset) -> list:
        def compute_average_column(column):
            """
            Function compute_average_list
            """
            if column.type == MDAVType.continous:
                avg_value = compute_average_continous(column.record)
            elif column.type == MDAVType.ordinal:
                avg_value = compute_average_ordinal(column.record)
            elif column.type == MDAVType.nominal:
                avg_value = compute_average_nominal(column.record)
            elif column.type == MDAVType.ignore:
                avg_value = ''  # This should not happen

            return avg_value

        def compute_average_continous(list):
            return mean(list)

        def compute_average_ordinal(list):
            # TODO: Use convex median
            return median_low(list)

        def compute_average_nominal(list):
            c = Counter(list)
            return c.most_common()[0][0]

        return [
            compute_average_column(column)
            for column in columns
        ]

    def find_most_distant_record(dataset: Dataset, x: list) -> list:
        """
        Function find_most_distant_record
        """
        max_distance = 0
        for y in dataset.records:
            distance = get_distance(x, y)
            if distance > max_distance:
                max_element, max_distance = y, distance
        return max_element

    def extract_cluster(dataset: Dataset,
                        x: list,
                        max_elements: int) -> (Dataset, Dataset):
        """
        Function extract_cluster
        """
        sorted_records = sorted(
            dataset.records,
            key=lambda record: get_distance(record, x)
        )
        new_cluster = Dataset(sorted_records[:max_elements], dataset.types)
        dataset = Dataset(sorted_records[max_elements+1:], dataset.types)

        return new_cluster, dataset

    # endregion

    logger.info(f'Computing MDAC generic for {len(dataset.records)} rows')

    private_data = []
    while len(dataset.records) >= 3 * k:
        logger.info(f'1. Iterating with {len(dataset.records)} records')

        columns, means, stdevs = compute_auxiliary_tables(dataset)

        # (1a) Compute the average record x̃ of all records in R. The average
        #      record is computed attribute-wise.
        x_avg = compute_average_record(dataset)
        logger.info(f'1a. Average record is {x_avg}')

        # (1b) Consider the most distant record xr to the average record x̃
        #      using an appropriate distance
        xr = find_most_distant_record(dataset, x_avg)
        logger.info(f'1b. Most distant record to average (xr) is {xr}')

        # (1c) Find the most distant record xs from the record xr considered in
        #      the previous step
        xs = find_most_distant_record(dataset, xr)
        logger.info(f'1c. Most distant record to average record (xs) is {xs}')

        # (1d) Form two clusters around xr and xs , respectively. One cluster
        #      contains xr and the k − 1 records closest to xr. The other
        #      cluster contains xs and the k − 1 records closest to xs.
        # (1e) Take as a new dataset R the previous dataset R minus the
        #      clusters formed around xr and xs in the last instance of
        #      Step 1d.
        cluster_xr, dataset = extract_cluster(dataset, xr, k - 1)
        columns, means, stdevs = compute_auxiliary_tables(cluster_xr)
        cluster_xr_avg = compute_average_record(cluster_xr)
        logger.info(f'1d. Cluster xr average record is {xr}')
        private_data.append(cluster_xr_avg)

        cluster_xs, dataset = extract_cluster(dataset, xs, k - 1)
        columns, means, stdevs = compute_auxiliary_tables(cluster_xs)
        cluster_xs_avg = compute_average_record(cluster_xs)
        logger.info(f'1d. Cluster xs average record is {xs}')
        private_data.append(cluster_xs_avg)
        logger.info(f'1e. New dataset size is {len(private_data)}')

    if len(dataset.records) >= 2 * k:

        logger.info(
            f'2. Entering step 2 with {len(dataset.records)} records (>=2k)'
        )
        # (2a) Compute the average record x̃ of the remaining records in R
        x_avg = compute_average_record(dataset)
        logger.info(f'2a. Average record is {x_avg}')
        # (2b) Find the most distant record xr from x̃
        xr = find_most_distant_record(dataset, x_avg)
        logger.info(f'2b. Most distant record to average (xr) is {xr}')
        # (2c) Form a cluster containing xr and the k − 1 records closest to xr
        cluster_xr, dataset = extract_cluster(dataset, xr, k - 1)
        columns, means, stdevs = compute_auxiliary_tables(cluster_xr)
        cluster_xr_avg = compute_average_record(cluster_xr)
        logger.info(f'2c. Cluster xr average record is {xr}')
        private_data.append(cluster_xr_avg)
        # (2d) form another cluster containing the rest of records

    else:

        logger.info(
            f'2. Entering step 2 with {len(dataset.records)} records (<2k)'
        )
        # else (less than 2k records in R) form a new cluster with the
        # remaining records

    columns, means, stdevs = compute_auxiliary_tables(dataset)
    cluster_rest = compute_average_record(dataset)
    logger.info(f'2/2d. Average record of rest of dataset is {cluster_rest}')
    private_data.append(cluster_rest)

    logger.info(f'2. New dataset size is {len(private_data)}')

    return private_data
