"""
Module helper functions
"""

import csv
from os.path import splitext

from babel.numbers import format_decimal, parse_decimal
from logzero import logger

from mdav.common import Column, Dataset, MDAVType


def opencsvfile(input_file, aux_file=None, headers=True,
                correct=True, delimiter=";") -> Dataset:
    """
    Loads a CSV file with data, and the auxiliary file with definitions
    """
    def check_and_convert_dataset(records, types, correct):
        new_records = []
        for record in records:
            new_record = []
            for i in range(len(record)):
                if types[i] in [MDAVType.continous, MDAVType.ordinal]:
                    try:
                        if types[i] == MDAVType.continous:
                            new_record.append(float(record[i]))
                        else:
                            new_record.append(int(record[i]))
                    except ValueError:
                        message = (f"Invalid value {record[i]} on "
                                   f"column {i+1}")
                        try:
                            if types[i] == MDAVType.continous:
                                new_record.append(
                                    float(parse_decimal(record[i],
                                                        locale="es_ES")))
                            else:
                                if correct:
                                    new_record.append(0)
                                    logger.debug(message)
                                else:
                                    raise ValueError(message)
                        except ValueError:
                            if correct:
                                new_record.append(0)
                                logger.debug(message)
                            else:
                                raise ValueError(message)
                elif types[i] == MDAVType.nominal:
                    new_record.append(record[i])
            new_records.append(new_record)
        new_types = [t for t in types if t != MDAVType.ignore]
        return Dataset(new_records, new_types)

    start_position = 1 if headers else 0
    logger.debug(f"Read CSV file {input_file}")
    if headers:
        logger.debug(f"Assuming files with headers, discarding first row")
    with open(input_file, "r") as f:
        reader = csv.reader(f, delimiter=delimiter)
        records = list(reader)
    records = records[start_position:]

    if aux_file is None:
        input_file_name, input_file_extension = splitext(input_file)
        aux_file = f'{input_file_name}-aux{input_file_extension}'

    logger.debug(f"Read CSV file {aux_file}")
    with open(aux_file, "r") as f:
        reader = csv.reader(f, delimiter=delimiter)
        types = list(reader)
    types = types[start_position:][0]

    try:
        types = [
            MDAVType(i) if i.isdigit() else MDAVType[i]
            for i in types
        ]
    except ValueError:
        raise ValueError(f'Unknown type on auxiliary file')

    dataset = check_and_convert_dataset(records, types, correct)

    return dataset


def localize_floats(row):
    return [
        str(el).replace('.', ',') if isinstance(el, float) else el
        for el in row
    ]


def savecsvfile(records, output_file, delimiter=";"):
    """
    Save a CSV file with a generated dataset
    """
    format_numbers = [
        [format_decimal(c, locale='es_ES') if isinstance(c, float) else c
         for c in r]
        for r in records
    ]
    records_as_str = [
        ['' if c is None else str(c) for c in r]
        for r in format_numbers
    ]
    with open(output_file, "w") as f:
        writer = csv.writer(f, delimiter=delimiter)
        writer.writerows(records_as_str)


def get_columns(dataset):
    columns = []
    for i in range(len(dataset.records[0])):
        column_record = []
        for record in dataset.records:
            column_record += [record[i]]
        columns.append(Column(column_record, dataset.types[i]))
    return columns
