#!/usr/bin/env python3
"""
Creates and runs some MDAV for TFG
"""

__author__ = "Javier Vilalta Bautista"
__version__ = "0.1.0"
__license__ = "MIT"


import logging
import argparse
import os
from os.path import splitext

from logzero import logger, logfile

from mdav import (calc_mdav_generic, get_dataset_columns_means,
                  get_dataset_columns_quantile50,
                  get_dataset_columns_quantile75,
                  get_dataset_columns_variances, get_dataset_mean_frequencies,
                  opencsvfile, savecsvfile)

DATA_FOLDER = 'data'


class RunMDAV:
    """ This class computes MDAV with sample files """

    # region Auxiliary methods

    def __init__(self):
        self.path = DATA_FOLDER
        self.testfile = ''
        self.comparefile = ''
        self.max_k = 2
        self.metric_aux_file = None

    def set_datafolder(self, folder):
        self.path = folder

    def set_testfile(self, file):
        self.testfile = os.path.join(self.path, file)

    def set_comparefile(self, file):
        self.comparefile = os.path.join(self.path, file)

    def set_max_k(self, new_max_k):
        self.max_k = new_max_k

    def set_metric_aux_file(self, aux_file):
        input_file_name, input_file_extension = splitext(aux_file)
        self.metric_aux_file = f'{input_file_name}-aux{input_file_extension}'

    # endregion

    # region Files to create

    def run_mdav_generic_files(self):
        self.run_mdav_generic_sample1()
        self.run_mdav_generic_US_census_1910()
        self.run_mdav_generic_US_census_1900()

    def run_mdav_generic_sample1(self):
        self.mdav_generic('sample1.csv')

    def run_mdav_generic_US_census_1910(self):
        self.mdav_generic('ed007-pg001a.csv')

    def run_mdav_generic_US_census_1900(self):
        self.mdav_generic('ed009-pg019a.csv')

    def run_mdav_generic_continous(self):
        self.mdav_generic('continous.csv')

    def run_mdav_generic_ordinal_dispersed(self):
        self.mdav_generic('ordinal-dispersed.csv')

    def run_mdav_generic_ordinal_fullvalues(self):
        self.mdav_generic('ordinal-allvalues.csv')

    def run_mdav_generic_nominal(self):
        self.mdav_generic('nominal.csv')

    # endregion

    # region Compute functions

    def mdav_generic(self, testfile):
        self.set_testfile(testfile)
        logger.info(f"Generating metrics files for {self.testfile}")
        self.metric_aux_file = None
        self.calc_metrics(self.testfile)
        for k in range(3, self.max_k+1):
            file, extension = os.path.splitext(testfile)
            self.set_comparefile(f'{file}-mdav-generic-{k}{extension}')
            dataset = opencsvfile(self.testfile)
            private_dataset = calc_mdav_generic(dataset, k)
            logger.info(f"Generating result file {self.comparefile}"
                        f" for {self.testfile}")
            savecsvfile(private_dataset, self.comparefile)
            logger.info(f"Generating metrics files for {self.comparefile}")
            self.set_metric_aux_file(self.testfile)
            self.calc_metrics(self.comparefile)

    # endregion

    # region Functions for metrics

    def calc_metrics(self, filename):
        self.calc_mean(filename)
        self.calc_variance(filename)
        self.calc_quantile50(filename)
        self.calc_quantile75(filename)
        self.calc_frequency(filename)

    def calc_mean(self, filename):
        self.calc_metric('mean', filename, get_dataset_columns_means)

    def calc_variance(self, filename):
        self.calc_metric('variance', filename, get_dataset_columns_variances)

    def calc_quantile50(self, filename):
        self.calc_metric('quantile50', filename,
                         get_dataset_columns_quantile50)

    def calc_quantile75(self, filename):
        self.calc_metric('quantile75', filename,
                         get_dataset_columns_quantile75)

    def calc_frequency(self, filename):
        self.calc_metric('frequency', filename, get_dataset_mean_frequencies)

    def calc_metric(self, metric_name, testfile, metric_function):
        file, extension = os.path.splitext(testfile)
        testfile = testfile
        comparefile = f'{file}-{metric_name}{extension}'
        result = [metric_function(testfile, aux_file=self.metric_aux_file)]

        logger.info(f"Generating {metric_name} file {comparefile}"
                    f" for {testfile}")
        savecsvfile(result, comparefile)

    # endregion


def main(args):
    """ Main entry point of the app """
    logfile("mdav-generic-logfile.log", backupCount=3)
    if args.verbosity == 0:
        logger.setLevel(logging.ERROR)
    elif args.verbosity == 1:
        logger.setLevel(logging.INFO)
    elif args.verbosity == 2:
        logger.setLevel(logging.WARNING)
    elif args.verbosity == 3:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.ERROR)
    logger.info("Starting MDAV")
    logger.info(args)
    runmdav = RunMDAV()
    runmdav.set_max_k(args.parameter)
    runmdav.set_datafolder(args.folder_origin)
    runmdav.run_mdav_generic_files()


if __name__ == "__main__":
    """ This is executed when run from the command line """
    parser = argparse.ArgumentParser(
        prog="MDAV-TFG",
        description="Calculate MDAV files for TFG"
    )

    # Optional argument for parameter
    parser.add_argument(
        "-k", action="store", dest="parameter",
        type=int, default=4,
        help="Specificies parameter k for MDAV"
    )

    # Optional argument for input file
    parser.add_argument(
        "-f", "--folder_origin", action="store",
        dest="folder_origin", default=DATA_FOLDER
    )

    # Optional verbosity counter (eg. -v, -vv, -vvv, etc.)
    parser.add_argument(
        "-v", "--verbose",
        action="count",
        dest="verbosity",
        default=0,
        help="Verbosity (-v, -vv, etc)")

    # Specify output of "--version"
    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s (version {version})".format(version=__version__))

    args = parser.parse_args()
    main(args)
